#include"dumberdowner.h"

#include<stdlib.h>
#include<assert.h>

// This is the dumbing down pass.
//
// Complex expressions are to be "broken down" into simpler ones until the AST
// can be trivially translated to the target architecture.
//
// This file along with CG is strictly for IA-32 and will fail for other
// architecture.

#include"x86.h"
#include<string.h>

/* Split away complex expression into a new local variable */
static AST *varify(AST *tlc, AST *chunk, AST *stmtPrev, AST *stmt, AST *e) {
	static size_t vidx = 0;
	
	VarTableEntry *vte = calloc(1, sizeof(*vte));
	vte->kind = VARTABLEENTRY_VAR;
	vte->type = e->expression.type;
	vte->data.var.color = -1;
	vte->data.var.precolored = false;
	vte->data.var.degree = 0;
	vte->data.var.priority = 0;
	vte->data.var.reachingDefs = NULL;
	vte->data.var.name = malp("$dumb%lu", vidx++);
	
	// Add to var array
	tlc->chunk.vars = realloc(tlc->chunk.vars, sizeof(*tlc->chunk.vars) * (++tlc->chunk.varCount));
	tlc->chunk.vars[tlc->chunk.varCount - 1] = vte;
	
	// Alter AST
	
	ASTExprVar *ev[2];
	for(int i = 0; i < 2; i++) {
		ev[i] = malloc(sizeof(ASTExprVar));
		ev[i]->nodeKind = AST_EXPR_VAR;
		ev[i]->type = e->expression.type;
		ev[i]->thing = vte;
	}
	
	ASTStmtAssign *assign = malloc(sizeof(*assign));
	assign->nodeKind = AST_STMT_ASSIGN;
	assign->what = (AST*) ev[0];
	assign->to = e;
	
	if(stmtPrev) {
		stmtPrev->statement.next = (AST*) assign;
	} else {
		chunk->chunk.statementFirst = (AST*) assign;
	}
	assign->next = stmt;
	
	// Reset ud-chain
	
	vte->data.var.usedefFirst = NULL;
	vte->data.var.usedefLast = NULL;
	
	return (AST*) ev[1];
}

static AST *xopify(AST *tlc, AST *chunk, AST *stmtPrev, AST *stmt, AST *e) {
	return varify(tlc, chunk, stmtPrev, stmt, e);
}

static bool is_incomplete_mul(AST *stmt) {
	if(stmt->nodeKind != AST_STMT_ASSIGN) {
		return false;
	}
	
	if(stmt->stmtAssign.what->nodeKind != AST_EXPR_VAR) {
		
	}
	
	return true;
}

struct DumbenState {
	int effective;
};

static void dumben_visitor(AST **nptr, AST *stmt, AST *stmtPrev, AST *chu, AST *tlc, void *ud) {
	struct DumbenState *this = ud;
	
	AST *s = *nptr;
	
	if(s->nodeKind == AST_STMT_IF) {
		
		AST *e = s->stmtIf.expression;
		
		if(e->nodeKind == AST_EXPR_BINARY_OP && binop_is_comparison(e->exprBinOp.operator)) {
		
			if(is_xop(e->exprBinOp.operands[0]) == XOP_NOT_XOP) {
				e->exprBinOp.operands[0] = xopify(tlc, chu, stmtPrev, s, e->exprBinOp.operands[0]);
				this->effective = 1;
			}
			
			if(is_xop(e->exprBinOp.operands[1]) == XOP_NOT_XOP) {
				e->exprBinOp.operands[1] = xopify(tlc, chu, stmtPrev, s, e->exprBinOp.operands[1]);
				this->effective = 1;
			}
			
			if(is_xop(e->exprBinOp.operands[0]) == XOP_MEM && is_xop(e->exprBinOp.operands[1]) == XOP_MEM) {
				// Can't have two mems; put one in var
				
				e->exprBinOp.operands[1] = varify(tlc, chu, stmtPrev, s, e->exprBinOp.operands[1]);
				this->effective = 1;
			}
			
		} else {
			
			s->stmtIf.expression = varify(tlc, chu, stmtPrev, s, e);
			this->effective = 1;
			
		}
		
	} else if(s->nodeKind == AST_STMT_LOOP) {
		
	} else if(s->nodeKind == AST_STMT_RETURN) {
		
		// ret specifically returns in eax always, so it needs to be in a var
		if(s->stmtReturn.val && (!is_xop(s->stmtReturn.val) || s->stmtReturn.val->nodeKind == AST_EXPR_PRIMITIVE)) {
			
			s->stmtReturn.val = varify(tlc, chu, stmtPrev, s, s->stmtReturn.val);
			this->effective = 1;
			
		}
		
	} else if(s->nodeKind == AST_STMT_EXPR && s->stmtExpr.expr->nodeKind == AST_EXPR_CALL) {
		
		// Turn this:
		//  f(x);
		// into:
		//  a = f(x);
		
		// So `a` would be colored as `eax`
		
		s->stmtExpr.expr = varify(tlc, chu, stmtPrev, s, s->stmtExpr.expr);
		
		// Purge this statement entirely, otherwise it'd be
		//  a = f(x);
		//  a;
		if(stmtPrev) {
			assert(stmtPrev->statement.next->statement.next == s);
			stmtPrev->statement.next->statement.next = s->statement.next;
		} else {
			assert(chu->chunk.statementFirst->statement.next = s);
			chu->chunk.statementFirst->statement.next = s->statement.next;
		}
		
	} else if(s->nodeKind == AST_STMT_ASSIGN) {
		
		if(ast_expression_equal(s->stmtAssign.what, s->stmtAssign.to) && stmtPrev) {
			// NOP; remove this statement from AST
			stmtPrev->statement.next = s->statement.next;
			this->effective = 1;
		} else {
			
			if(s->stmtAssign.what->nodeKind == AST_EXPR_UNARY_OP && s->stmtAssign.what->exprUnOp.operator == UNOP_DEREF
				&& s->stmtAssign.to->nodeKind == AST_EXPR_UNARY_OP && s->stmtAssign.to->exprUnOp.operator == UNOP_DEREF) {
				
				s->stmtAssign.to = varify(tlc, chu, stmtPrev, s, s->stmtAssign.to);
				this->effective = 1;
			} else if(s->stmtAssign.what && s->stmtAssign.what->nodeKind == AST_EXPR_VAR && s->stmtAssign.what->exprVar.thing->kind == VARTABLEENTRY_VAR && s->stmtAssign.to->nodeKind == AST_EXPR_CALL) {
				ASTExprCall *call = &s->stmtAssign.to->exprCall;
				
				int argCount = call->what->expression.type->function.argCount;
				
				for(int i = 0; i < argCount; i++) {
					if(is_xop(call->args[i]) == XOP_NOT_XOP) {
						call->args[i] = xopify(tlc, chu, stmtPrev, s, call->args[i]);
						this->effective = 1;
					}
				}
			} else if(s->stmtAssign.to && s->stmtAssign.to->nodeKind == AST_EXPR_UNARY_OP && s->stmtAssign.to->exprUnOp.operator == UNOP_NEGATE && !ast_expression_equal(s->stmtAssign.what, s->stmtAssign.to->exprUnOp.operand)) {
				// Turn this:
				//  a = -b
				// into
				//  a = b
				//  a = -a
				
				AST *ev[2] = {
					ast_deep_copy(s->stmtAssign.what),
					ast_deep_copy(s->stmtAssign.what)
				};
				
				AST *negation = s->stmtAssign.to;
				
				s->stmtAssign.to = negation->exprUnOp.operand;
				negation->exprUnOp.operand = ev[0];
				
				AST *assign2 = malloc(sizeof(ASTStmtAssign));
				assign2->nodeKind = AST_STMT_ASSIGN;
				assign2->stmtAssign.what = ev[1];
				assign2->stmtAssign.to = negation;
				
				assign2->statement.next = s->statement.next;
				s->statement.next = assign2;
				
				this->effective = 1;
			} else if(is_xop(s->stmtAssign.to) == XOP_NOT_XOP) {
				if(s->stmtAssign.to->nodeKind == AST_EXPR_UNARY_OP && s->stmtAssign.to->exprUnOp.operator == UNOP_DEREF) {
					s->stmtAssign.to->exprUnOp.operand = varify(tlc, chu, stmtPrev, s, s->stmtAssign.to->exprUnOp.operand);
					
					this->effective = 1;
				} else if(s->stmtAssign.to->nodeKind == AST_EXPR_BINARY_OP && s->stmtAssign.to->exprBinOp.operator == BINOP_MUL) {
					
					if(s->stmtAssign.what->nodeKind != AST_EXPR_VAR) {
						s->stmtAssign.what = varify(tlc, chu, stmtPrev, s, s->stmtAssign.what);
						this->effective = 1;
					}
					
					/*if(!(s->statement.next && s->statement.next->nodeKind == AST_STMT_ASSIGN && s->statement.next->stmtAssign.)) {
						
					}*/
					
					// TODO: implement multiplication
					
				} else if(s->stmtAssign.to->nodeKind == AST_EXPR_BINARY_OP && s->stmtAssign.to->exprBinOp.operator < BINOP_SIMPLES && !ast_expression_equal(s->stmtAssign.what, s->stmtAssign.to->exprBinOp.operands[0])) {
					// Turn this:
					//  a = b op c
					// into
					//  a = b
					//  a = a op c
					
					AST *assign2 = malloc(sizeof(ASTStmtAssign));
					assign2->nodeKind = AST_STMT_ASSIGN;
					assign2->stmtAssign.what = ast_deep_copy(s->stmtAssign.what);
					assign2->stmtAssign.to = s->stmtAssign.to->exprBinOp.operands[0];
					
					if(stmtPrev) {
						stmtPrev->statement.next = assign2;
					} else {
						chu->chunk.statementFirst = assign2;
					}
					assign2->statement.next = s;
					
					s->stmtAssign.to->exprBinOp.operands[0] = ast_deep_copy(s->stmtAssign.what);
					
					this->effective = 1;
				} else if(s->stmtAssign.to->nodeKind == AST_EXPR_BINARY_OP && !is_xop(s->stmtAssign.to->exprBinOp.operands[0])) {
					s->stmtAssign.to->exprBinOp.operands[0] = xopify(tlc, chu, stmtPrev, s, s->stmtAssign.to->exprBinOp.operands[0]);
					this->effective = 1;
				} else if(s->stmtAssign.to->nodeKind == AST_EXPR_BINARY_OP && !is_xop(s->stmtAssign.to->exprBinOp.operands[1])) {
					s->stmtAssign.to->exprBinOp.operands[1] = xopify(tlc, chu, stmtPrev, s, s->stmtAssign.to->exprBinOp.operands[1]);
					this->effective = 1;
				}
			}
		}
		
	}
}

static void pre_dumb_visitor(AST **nptr, AST *stmt, AST *stmtPrev, AST *chunk, AST *tlc, void *ud) {
	AST *n = *nptr;
	
	if(n == tlc) {
		if(tlc->chunk.functionType) {
			size_t argCount = n->chunk.functionType->function.argCount;
			
			// First argCount vtes in list are the arguments
			for(size_t i = 0; i < argCount; i++) {
				VarTableEntry *vte = n->chunk.vars[i];
				
				ASTExprStackPointer *stck = calloc(1, sizeof(*stck));
				stck->nodeKind = AST_EXPR_STACK_POINTER;
				stck->type = type_pointer_wrap(vte->type);
				
				ASTExprPrimitive *offset = calloc(1, sizeof(*offset));
				offset->nodeKind = AST_EXPR_PRIMITIVE;
				offset->type = primitive_parse("u32");
				offset->val = 4 + i * 4;
				
				ASTExprBinaryOp *sum = calloc(1, sizeof(*sum));
				sum->nodeKind = AST_EXPR_BINARY_OP;
				sum->type = stck->type;
				sum->operands[0] = (AST*) stck;
				sum->operands[1] = (AST*) offset;
				sum->operator = BINOP_ADD;
				
				ASTExprUnaryOp *deref = calloc(1, sizeof(*deref));
				deref->nodeKind = AST_EXPR_UNARY_OP;
				deref->type = vte->type;
				deref->operand = (AST*) sum;
				deref->operator = UNOP_DEREF;
				
				ASTExprVar *evar = calloc(1, sizeof(*evar));
				evar->nodeKind = AST_EXPR_VAR;
				evar->type = vte->type;
				evar->thing = vte;
				
				ASTStmtAssign *ass = calloc(1, sizeof(*ass));
				ass->nodeKind = AST_STMT_ASSIGN;
				ass->next = NULL;
				ass->what = (AST*) evar;
				ass->to = (AST*) deref;
				ass->next = n->chunk.statementFirst;
				
				if(n->chunk.statementFirst) {
					n->chunk.statementFirst = ass;
				} else {
					assert(!n->chunk.statementLast);
					n->chunk.statementFirst = n->chunk.statementLast = ass;
				}
			}
		}
	}
}

void dumben_pre(AST *tlc) {
	generic_visitor(&tlc, NULL, NULL, tlc, tlc, NULL, pre_dumb_visitor);
}

void dumben_go(AST* tlc) {
	size_t i = 0;
	while(1) {
		fprintf(stderr, "; Dumbing down %lu...\n", i++);
		
		struct DumbenState state;
		memset(&state, 0, sizeof(state));
		
		generic_visitor(&tlc, NULL, NULL, tlc, tlc, &state, dumben_visitor);
		
		int successful = state.effective;
		
		if(!successful) {
			break;
		}
	}
}
